
#include <cmath>
#include <iostream>
#include <fstream>
#include <string>

using namespace std;

int node_id;
int element_id;
int slice_size;

#define SIZEX 64
#define SIZEY 64
#define SIZEZ 15


  // --------------------------------------------------------------------------
  // Write a node.
void w_node( ofstream* outgeo, ofstream* output, double x, double y, double z )
{
  ++ node_id;
  
  (*output) << node_id << ", " << x << ", " << y << ", " << z << endl;
  
  (*outgeo) << "ND," << node_id <<"," << x << ", " << y << ", " << z << endl;
}


  // --------------------------------------------------------------------------
  // Write an element.
void w_element( ofstream* outgeo, ofstream* output, int x, int y, int z )
{
  ++ element_id;
  
  int n_slice_size = (SIZEX+1) * (SIZEY+1);
  
    // Nodes of the lower layer.
  int n1 = z * n_slice_size + y * (SIZEX+1) + x + 1;
  int n2 = z * n_slice_size + y * (SIZEX+1) + x + 1 + 1;
  int n3 = z * n_slice_size + ( y + 1 ) * (SIZEX+1) + x + 1 + 1;
  int n4 = z * n_slice_size + ( y + 1 ) * (SIZEX+1) + x + 1;
  
    // Nodes of the upper layer.
  int n5 = ( z + 1 ) * n_slice_size + y * (SIZEX+1) + x + 1;
  int n6 = ( z + 1 ) * n_slice_size + y * (SIZEX+1) + x + 1 + 1;
  int n7 = ( z + 1 ) * n_slice_size + ( y + 1 ) * (SIZEX+1) + x + 1 + 1;
  int n8 = ( z + 1 ) * n_slice_size + ( y + 1 ) * (SIZEX+1) + x + 1;
  
    // Output for .inp file.
  (*output) << element_id << ", " << n1 << ", " << n2 << ", " << n3 << ", "
            << n4 << ", " << n5 << ", " << n6 << ", " << n7 << ", " << n8
			<< endl;
  
    // Output for .geo file.
  (*outgeo) << "EL" << "," << element_id << ",VL,0,8, " << n1 << ", " << n2 << ", " << n3 << ", "
            << n4 << ", " << n5 << ", " << n6 << ", " << n7 << ", " << n8
			<< endl;
}


  // --------------------------------------------------------------------------
  // Main program.
int main( int argc, char** argv )
{
  cerr << endl;
  
    // Process command line arguments.
  if ( argc != 4 )
  {
    cerr << "Usage: " << argv[ 0 ] << " intput.txt output.inp output.geo"
	     << endl;
	return 1;
  }
  char *input_file = argv[ 1 ];
  char *output_file = argv[ 2 ];
  char *geo_file = argv[ 3 ];
  
  node_id = 0;
  element_id = 0;
  
    // Open input file.
  ifstream *input = new ifstream;
  input->open( input_file );
  
    // Count input data points.
  input->seekg ( 0, ios::beg );
  long input_size = 0;
  long bone_size = 0;
  char c = 0;
  while ( 1 )
  {
    (*input) >> c;
	
	if ( input->eof() )  // End of file reached.
	{
	  break;
	}
	
	if ( c == '0' || c == '1' )  // Data point read.
	{
	  ++ input_size;
	  if ( c == '1' )
	  {
	    ++ bone_size;
	  }
	}
  }
  cerr << "Input file size: " << input_size << endl;
  cerr << "      bone size: " << bone_size << endl;
  
    // Size of one data layer in the 3D volume.
  slice_size = SIZEX * SIZEY;
  if ( input_size < SIZEX * SIZEY * SIZEZ )
  {
    cerr << "ERROR: Wrong input size!";
	return 1;
  }
  
    // Reset input reader.
  input->close();
  delete input;
  input = new ifstream;
  input->open( input_file );
  
    // Read input data.
  int rid_bone = 0;
  int rid_fat = 0;
  cerr << "Reading input data..." << endl;
  
    // Allocate memory for data volume.
  char *data = new char[ input_size ];
  
    // Fill data buffer.
  int ii = 0;
  while ( 1 )
  {
    (*input) >> c;
	
	  // Quit conditions: end of file / end of data buffer.
	if ( input->eof() || ii >= input_size )
	{
	  if ( ii < input_size )
	  {
	    cerr << "ERROR: Wrong input size" << endl;
		return 1;
	  }
	  break;
	}
	
    if ( c == '0' )  // Fat.
	{
	  data[ ii ] = 0;
	  ++ ii;
	  rid_fat ++;
	}
	if ( c == '1' )  // Bone.
	{
	  data[ ii ] = 1;
	  ++ ii;
	  rid_bone ++;
	}
  }
  cerr << "  " << rid_bone << " bone elements read." << endl;
  cerr << "  " << rid_fat << " fat elements read." << endl;
  cerr << endl;
  
    // Close input file.
  input->close();
  delete input;
  
  
  // OUTPUT *******************************************************************
  
    // Open output file.
  ofstream *output = new ofstream;
  output->open( output_file );
  
    // Open geo file.
  ofstream *outgeo = new ofstream;
  outgeo->open( geo_file );
  
    // Nodes.
  (*output) << endl;
  (*output) << "*NODE, NSET=Nall" << endl
            << endl;
  cerr << "Writing nodes..." << endl;
  int n_sizex = SIZEX + 1;
  int n_sizey = SIZEY + 1;
  int n_sizez = SIZEZ + 1;
  int n_slice_size = n_sizex * n_sizey;
  int n_input_size = n_sizex * n_sizey * n_sizez;
  
    // Iterate through nodes of the mesh.
  for ( int z = 0; z < n_sizez; ++ z )
  {
    for ( int y = 0; y < n_sizey; ++ y )
	{
	  for ( int x = 0; x < n_sizex; ++ x )
	  {
	    w_node( outgeo, output, x, y, z );
	  }
	}
  }
  
  cerr << "  done." << endl;
  (*output) << endl;
  
    // Bone material definition for .geo file.
  (*outgeo) << endl;
  (*outgeo) << "EGROUP,1,SOLID,0,1,0,0,0,0,0,0" << endl;
  (*outgeo) << "PICK_MAT,1,BONE,MKS" << endl;
  (*outgeo) << "RCONST,1,1,1,9,0,0,0,0,0,0,0,0,0" << endl;
  (*outgeo) << endl;
  
    // Bone elements.
  int n_bone = 0;
  cerr << "Writing bone elments... ";
  (*output) << "*ELEMENT, TYPE=C3D8R, ELSET=Bone" << endl;
  
  for ( int z = 0; z < SIZEZ; ++ z )
  {
    for ( int y = 0; y < SIZEY; ++ y )
	{
	  for ( int x = 0; x < SIZEX; ++ x )
	  {
	    if ( data[ z * slice_size + y * SIZEX + x ] == 1 )
		{
		  w_element( outgeo, output, x, y, z );
		  ++ n_bone;
		}
	  }
	}
  }
  
  cerr << "Done." << endl << "  " << n_bone << " bone elements written." << endl;
  (*output) << endl;
  
    // Define inter-bone material.
  (*outgeo) << endl;
  (*outgeo) << "EGROUP,2,SOLID,0,1,0,0,0,0,0,0" << endl;
  (*outgeo) << "PICK_MAT,2,RUBBER,MKS" << endl;
  (*outgeo) << "RCONST,2,1,1,9,0,0,0,0,0,0,0,0,0" << endl;
  (*outgeo) << endl;
  
    // Liquid elements.
  int n_liq = 0;
  cerr << "Writing fat elements... ";
  (*output) << "*ELEMENT, TYPE=C3D8R, ELSET=Fat" << endl;
  
  for ( int z = 0; z < SIZEZ; ++ z )
  {
    for ( int y = 0; y < SIZEY; ++ y )
	{
	  for ( int x = 0; x < SIZEX; ++ x )
	  {
	    if ( data[ z * slice_size + y * SIZEX + x ] == 0 )
		{
		  w_element( outgeo, output, x, y, z );
		  ++ n_liq;
		}
	  }
	}
  }
  
  cerr << "Done." << endl << "  " << n_liq << " fat elements written." << endl;
  (*output) << endl;
  
    // Fixed elements.
  (*output) << "*NSET,NSET=Nfix" << endl;
  for ( int i = 0; i < (SIZEX+1) * (SIZEY+1); ++ i )
  {
    (*output) << i+1 << ", " << endl;
  }
  (*output) << endl;
  
  (*output) << "*BOUNDARY" << endl
            << "Nfix,1,3" << endl
	        << endl;
    // Material properties.
  (*output) << "*MATERIAL,NAME=MBone" << endl
            << "*ELASTIC" << endl
            << "210000.,.3" << endl
            << "*DENSITY" << endl
            << "7.8e-9" << endl
            << "*SOLID SECTION,ELSET=Bone,MATERIAL=MBone" << endl
            << endl;
   (*output) << "*MATERIAL,NAME=MFat" << endl
            << "*ELASTIC" << endl
            << "21000.,.3" << endl
            << "*DENSITY" << endl
            << "7.8e-9" << endl
            << "*SOLID SECTION,ELSET=Fat,MATERIAL=MFat" << endl
            << endl;
  
    // Loaded elements.
  (*output) << "**" << endl
            << "*STEP" << endl
            << "*STATIC" << endl
            << "*CLOAD" << endl;
            // << "8,3,-1. " << endl
  for ( int i = 0; i < (SIZEX+1) * (SIZEY+1); ++ i )
  {
    (*output) << ( i + ( SIZEX + 1 ) * ( SIZEY + 1 ) * ( SIZEZ ) + 1 ) << ",3,-1. " << endl;
  }
  (*output) << endl;

			
  (*output) << "*NODE PRINT,NSET=Nall" << endl
            << "U" << endl
			<< endl;
			
  (*output) << "*EL PRINT,ELSET=Bone" << endl
            << "S" << endl
			<< endl;
			
  (*output) << "*EL PRINT,ELSET=Fat" << endl
            << "S" << endl
			<< endl;
			
  (*output) << "*END STEP" << endl
			<< endl;
  
  (*outgeo) << endl;
  (*outgeo) << "DND," << 1 << ",AL,0," << (SIZEX+1) * (SIZEY+1) << ",1" << endl;
  (*outgeo) << "FND," << ( SIZEX + 1 ) * ( SIZEY + 1 ) * ( SIZEZ ) + 1
            << ",FZ,1," << ( SIZEX + 1 ) * ( SIZEY + 1 ) * ( SIZEZ + 1 )
			<< ",1" << endl;
  (*outgeo) << "ACTECLR,1,EG,1" << endl;
  (*outgeo) << "EVAL_BOUND,1,0,20,1" << endl;
  
    // Close output file.
  output->close();
  delete output;
  
  outgeo->close();
  delete outgeo;
  
  delete data;
  return 0;
}
